/* This file is part of rpipe
   Copyright (C) 2018-2021 Sergey Poznyakoff
   License GPLv3+: GNU GPL version 3 or later
   <http://gnu.org/licenses/gpl.html>

   This is free software: you are free to change and redistribute it.
   There is NO WARRANTY, to the extent permitted by law.
*/
#include <config.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <sysexits.h>
#include <unistd.h>
#include <poll.h>
#include <limits.h>
#include "rpipe.h"
#include "iobuf.h"

void
client_sig_handler(int sig)
{
	error(EX_OSERR, 0, "got signal %s", strsignal(sig));
}

static int client_sigv[] = {
	SIGCHLD,
	SIGTERM,
	SIGQUIT,
	SIGINT,
	SIGHUP,
	SIGUSR1,
	SIGALRM,
	SIGPIPE
};
static int client_sigc = sizeof(client_sigv)/sizeof(client_sigv[0]);

struct rpipe_client {
	struct iobuf ibuf;        /* Input buffer */
	struct iobuf obuf;        /* Output buffer */
};

static int
valid_pollfd(struct pollfd *pf, int n)
{
	while (n--) {
		if (pf->fd >= 0 && pf->events)
			return 1;
		++pf;
	}
	return 0;
}

static void
waitreply(int fd)
{
	enum {
		NET,
		IN,
		OUT,
		NFD
	};
	struct pollfd fds[NFD];
	struct rpipe_client client[NFD];
	struct crlfdot_decoder_state state = CRLFDOT_DECODER_STATE_INIT;
	memset(&fds, 0, sizeof(fds));

	fds[NET].fd = fd;
	fds[NET].events = POLLIN;

	fds[IN].fd = 0;
	fds[IN].events = POLLIN;

	fds[OUT].fd = 1;
	fds[OUT].events = POLLOUT;

	memset(&client, 0, sizeof(client));

	while (valid_pollfd(fds, NFD)) {
		ssize_t n = poll(fds, NFD, -1);
		if (n == -1) {
			if (errno != EINTR)
				error(0, errno, "poll");
			continue;
		}

		if (fds[NET].revents & POLLIN) {
			n = iobuf_fill(&client[NET].ibuf, fds[NET].fd);
			if (n == -1) {
				error(EX_OSERR, errno, "read");
			} else if (n == 0) {
				fds[NET].events = 0;
			} else {
				fds[OUT].events |= POLLOUT;
				if (iobuf_avail_size(&client[NET].ibuf) == 0)
					fds[NET].events &= ~POLLIN;
			}
		}
		if (fds[NET].revents & POLLOUT) {
			iobuf_transfer(&client[NET].obuf, &client[IN].ibuf);

			n = iobuf_flush(&client[NET].obuf, fds[NET].fd);
			if (n == -1) {
				error(EX_OSERR, errno, "write");
			} else if (n == 0) {
				error(EX_OSERR, ENOSPC, "write");
			}
			if (iobuf_data_size(&client[NET].obuf) == 0) {
				fds[NET].events &= ~POLLOUT;
				if (fds[IN].events == 0 &&
				    iobuf_data_size(&client[IN].ibuf) == 0) {
					shutdown(fds[NET].fd, SHUT_WR);
				}
			}
		}

		if ((fds[IN].revents & POLLIN) == 0 ||
		    (n = iobuf_fill(&client[IN].ibuf, fds[IN].fd)) == 0) {
			if (iobuf_data_size(&client[NET].obuf) == 0 &&
			    iobuf_data_size(&client[IN].ibuf) == 0) {
				shutdown(fds[NET].fd, SHUT_WR);
				fds[NET].events &= ~POLLOUT;
			}
			fds[IN].events = 0;
		} else if (n == -1) {
			error(EX_OSERR, errno, "read");
		} else {
			fds[NET].events |= POLLOUT;
		}

		if (fds[OUT].revents & POLLOUT) {
			crlfdot_decoder(&state, &client[OUT].obuf,
					&client[NET].ibuf);
			if (iobuf_avail_size(&client[NET].ibuf))
				fds[NET].events |= POLLIN;
			if (iobuf_data_size(&client[OUT].obuf) == 0)
				fds[OUT].events &= ~POLLOUT;
			else {
				n = iobuf_flush(&client[OUT].obuf, fds[OUT].fd);
				if (n == -1) {
					error(EX_OSERR, errno, "write");
				} else if (n == 0) {
					error(EX_OSERR, ENOSPC, "write");
				}
			}
			if (state.state == -1) {
				/* Final state reached. */
				fds[IN].events = 0;
			}
		}
	}

	if (state.state == -1 && iobuf_data_size(&client[NET].ibuf)) {
		char *p;
		unsigned long u = 0;
		static char dig[] = "0123456789";
		int c;
		int status;
		int ok = 0;
		while ((c = iobuf_getc(&client[NET].ibuf)) != -1
		       && (p = strchr(dig, c)) != NULL) {
			u = u * 10 + p - dig;
			ok = 1;
		}
		if (c != '\r')
			error(0, 0, "remote protocol error: CRLF missing");
		if (!ok || u > INT_MAX)
			error(EX_UNAVAILABLE, 0,
			      "remote protocol error: exit status out of range");
		status = u;
		if (verbose)
			report_prog_status(status, "remote command");
		if (WIFEXITED(status))
			exit(WEXITSTATUS(status));
	}
	exit(EX_OSERR);
}

static void
trim(char *buf)
{
	size_t len = strlen(buf);
	if (len > 0 && buf[len-1] == '\n') {
		--len;
		if (len > 0 && buf[len-1] == '\r')
			--len;
		buf[len] = 0;
	}
}

void
rpipe_client(struct addrinfo *ap, int argc, char **argv)
{
	int fd;
	FILE *fp;
	char buf[IO_BUF_SIZE];
	int i;

	signal_setup(client_sig_handler, client_sigv, client_sigc);

	fd = socket(ap->ai_addr->sa_family, SOCK_STREAM, 0);
	if (fd == -1)
		error(EX_OSERR, errno, "can't create socket");

	if (connect(fd, ap->ai_addr, ap->ai_addrlen))
		error(EX_OSERR, errno, "connect");

	fp = fdopen(fd, "w+");

	for (i = 0; i < argc; i++) {
		if (i > 0)
			fputc(' ', fp);
		if (argv[i][strcspn(argv[i], " \"\\")] == 0)
			fprintf(fp, "%s", argv[i]);
		else {
			char *p;

			fputc('"', fp);
			for (p = argv[i]; *p; p++) {
				if (*p == '"' || *p == '\\')
					fputc('\\', fp);
				fputc(*p, fp);
			}
			fputc('"', fp);
		}
	}
	fputc('\r', fp);
	fputc('\n', fp);
	if (ferror(fp))
		error(EX_OSERR, 0, "write error");

	if (!fgets(buf, sizeof(buf), fp))
		error(EX_OSERR, errno, "error reading response");
	trim(buf);
	if (buf[0] == '-')
		error(EX_UNAVAILABLE, 0, "%s", buf + 2);
	else if (buf[0] != '+')
		error(EX_UNAVAILABLE, 0, "unexpected response: %s", buf);

	waitreply(fd);
}

